import importlib.util
import os
import re
import sys
from logging import (
    INFO,
    NOTSET,
    Formatter,
    Handler,
    Logger,
    StreamHandler,
    getLogger,
)
from types import TracebackType
from typing import Optional, TypeVar

HandlerType = TypeVar("HandlerType", bound=Handler)


def _color_supported() -> bool:
    """Detection of color support."""
    if not importlib.util.find_spec("colorlog"):
        return False

    # NO_COLOR environment variable:
    if os.environ.get("NO_COLOR", None):
        return False

    if not hasattr(sys.stderr, "isatty") or not sys.stderr.isatty():
        return False
    else:
        return True


_default_handler: Optional[StreamHandler] = None
"""default root logger handler

if not configured, None
"""


def _get_root_logger_name() -> str:
    """get root logger name (library name)

    Returns
    -------
    str
        root logger name (library name)
    """
    return __name__.split(".")[0]


def create_default_formatter() -> Formatter:
    """create default formatter

    Returns
    -------
    Formatter
        default formatter
    """
    if _color_supported():
        from colorlog import ColoredFormatter

        return ColoredFormatter(
            "%(asctime)s - %(name)s:%(lineno)d%(log_color)s[%(levelname)s]%(reset)s - %(message)s"
        )
    else:
        return Formatter(
            "%(asctime)s - %(name)s:%(lineno)d[%(levelname)s] - %(message)s"
        )


default_formatter: Formatter = create_default_formatter()
"""default formatter"""


def get_handler(
    handler: HandlerType, formatter: Optional[Formatter] = None, level=NOTSET
) -> HandlerType:
    """configure handler in an easy api

    Parameters
    ----------
    handler : HANDLER

    formatter : Optional[Formatter], optional
        , by default None
    level : _type_, optional
        , by default NOTSET

    Returns
    -------
    HANDLER

    """
    handler.setLevel(level)
    handler.setFormatter(
        formatter if formatter else create_default_formatter()
    )
    return handler


def _create_default_handler() -> StreamHandler:
    return get_handler(StreamHandler())


def _configure_library_root_logger() -> None:
    global _default_handler

    if _default_handler:
        # This library has already configured the library root logger.
        return

    _default_handler = _create_default_handler()

    # Apply our default configuration to the library root logger.
    library_root_logger = get_root_logger()
    library_root_logger.addHandler(_default_handler)
    library_root_logger.setLevel(INFO)
    library_root_logger.propagate = False


def get_root_logger() -> Logger:
    """get library root logger of this package

    Returns
    -------
    Logger
        library root logger
    """
    _configure_library_root_logger()

    return getLogger(_get_root_logger_name())


def get_child_logger(name: str, propagate: bool = True) -> Logger:
    """get logger

    Parameters
    ----------
    name : str
        You shold assign '__name__'

    propagate : bool
        propagate to parent handler or not, by default True

    Returns
    -------
    Logger
        child logger

    Raises
    ------
    ValueError

    """
    root_logger = get_root_logger()

    _result_logger = re.match(rf"{_get_root_logger_name()}\.(.+)", name)
    if _result_logger:
        child_logger = root_logger.getChild(_result_logger.group(1))
    elif name == "__main__":
        child_logger = root_logger.getChild(name)
    else:
        raise ValueError("You should use '__name__'.")

    child_logger.propagate = propagate
    return child_logger


def enable_default_handler() -> None:
    """enable default handler"""
    _configure_library_root_logger()

    assert _default_handler is not None
    get_root_logger().addHandler(_default_handler)


def disable_default_handler() -> None:
    """disable default handler"""
    _configure_library_root_logger()

    assert _default_handler is not None
    get_root_logger().removeHandler(_default_handler)


class catch_default_handler:
    """catch default handler

    Example
    -------
    >>> _logger = get_child_logger(__name__)
    >>> with catch_default_handler():
    >>>    _logger.info("not log")
    >>> _logger.info("log")

    """

    def __enter__(self) -> None:
        disable_default_handler()

    def __exit__(
        self,
        exc_type: "Optional[type[Exception]]",
        exc_value: Optional[Exception],
        traceback: Optional[TracebackType],
    ) -> None:
        enable_default_handler()



import argparse
import os
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import Optional

from packing_packages.pack._core import packing_packages


def pack(args: argparse.Namespace) -> None:
    """pack conda environment"""
    if args.env_name is None:
        env_name = os.environ["CONDA_DEFAULT_ENV"]
    else:
        env_name = args.env_name

    dirpath_target = Path(args.dirpath_target).resolve()
    if not dirpath_target.is_dir():
        raise FileNotFoundError(dirpath_target)

    packing_packages(
        env_name=env_name,
        dirpath_target=dirpath_target,
        encoding=args.encoding,
        dry_run=args.dry_run,
    )


def add_arguments_pack(parser: argparse.ArgumentParser):
    parser.add_argument(
        "-n",
        "--env-name",
        type=str,
        default=None,
        help=(
            "Name of the conda environment to pack. If not specified, the current active environment is used."
        ),
    )
    parser.add_argument(
        "-d",
        "--dirpath-target",
        type=Path,
        default=".",
        help=(
            "Path to the directory where the packed environment will be saved. Defaults to the current directory."
        ),
    )
    parser.add_argument(
        "-D",
        "--dry-run",
        action="store_true",
        help=(
            "Perform a dry run without actually packing the environment. Useful for testing."
        ),
    )
    parser.add_argument(
        "-e",
        "--encoding",
        type=str,
        default=None,
        help=(
            "Encoding to use for subprocess execution. If not set, system default is used."
        ),
    )
    parser.set_defaults(func=pack)


def main(cli_args: Sequence[str], prog: Optional[str] = None) -> None:
    """Main function to parse arguments and call the packing function."""
    parser = argparse.ArgumentParser(
        prog=prog, description="pack conda environment"
    )
    add_arguments_pack(parser)
    args = parser.parse_args(cli_args)
    args.func(args)


if __name__ == "__main__":
    import sys

    main(sys.argv[1:])



import argparse
import os
from collections.abc import Sequence
from pathlib import Path
from typing import Optional

from packing_packages.install._core import install_packages


def install(args: argparse.Namespace) -> None:
    """Install packages in the conda environment."""
    if args.env_name is None:
        env_name = os.environ["CONDA_DEFAULT_ENV"]
    else:
        env_name = args.env_name

    dirpath_packages = Path(args.dirpath_packages).resolve()
    if not dirpath_packages.is_dir():
        raise FileNotFoundError(dirpath_packages)

    install_packages(
        dirpath_packages=dirpath_packages,
        env_name=env_name,
        encoding=args.encoding,
    )


def add_arguments_install(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "dirpath-packages",
        type=Path,
        default=".",
        help=(
            "Path to the directory containing packages to install (e.g., *.tar.bz2, *.conda). "
            "Defaults to the current directory ('.') if not specified."
        ),
        metavar="DIRPATH_PACKAGES",
    )
    parser.add_argument(
        "-n",
        "--env-name",
        type=str,
        default=None,
        help=(
            "Name of the conda environment where packages will be installed. "
            "If not specified, the currently active environment will be used."
        ),
    )
    parser.add_argument(
        "-e",
        "--encoding",
        type=str,
        default=None,
        help=(
            "Encoding to use for subprocess execution (e.g., when calling conda commands). "
            "Defaults to the system encoding if not specified."
        ),
    )
    parser.set_defaults(func=install)


def main(cli_args: Sequence[str], prog: Optional[str] = None) -> None:
    """Main function to parse arguments and call the install function."""
    parser = argparse.ArgumentParser(
        prog=prog, description="Install packages in the conda environment"
    )
    add_arguments_install(parser)
    args = parser.parse_args(cli_args)
    args.func(args)



import importlib.util
import inspect
import sys
from collections.abc import Callable, Iterable, Iterator
from typing import Any, Optional, TypeVar

from packing_packages.logging import get_child_logger

_logger = get_child_logger(__name__)

T = TypeVar("T")


def is_installed(package_name: str) -> bool:
    """Check if the package is installed.

    Parameters
    ----------
    package_name : str
        package name like `sklearn`

    Returns
    -------
    bool
        if installed, True
    """
    return bool(importlib.util.find_spec(package_name))


def is_argument(__callable: "Callable[..., Any]", arg_name: str) -> bool:
    """Check to see if it is included in the callable argument.

    Parameters
    ----------
    __callable : Callable

    arg_name : str
        argument name

    Returns
    -------
    bool
        if included, True
    """
    return arg_name in set(inspect.signature(__callable).parameters.keys())


class dummy_tqdm(Iterable[T]):
    """dummy class for 'tqdm'

    Parameters
    ----------
    __iterable : Iterable[T]
        iterable object
    """

    def __init__(self, __iterable: "Iterable[T]", *args, **kwargs) -> None:
        self.__iterable = __iterable

    def __iter__(self) -> "Iterator[T]":
        return iter(self.__iterable)

    def __getattr__(self, name: str) -> "Callable[..., None]":
        return self.__no_operation

    @staticmethod
    def __no_operation(*args, **kwargs) -> None:
        """no-operation"""
        return


def check_encoding(encoding: Optional[str]) -> str:
    """Check if the encoding is valid.

    Parameters
    ----------
    encoding : str
        encoding name

    Returns
    -------
    str
        encoding name
    """
    if encoding is None:
        encoding = sys.getdefaultencoding()
        _logger.info(f"Use default encoding: {encoding}.")
    try:
        "".encode(encoding)
        return encoding
    except LookupError as e:
        raise ValueError(f"Invalid encoding: {encoding}") from e



"""This package provides a set of functions to manage logging in Python applications.

It includes functions to enable or disable the default logging handler, catch the default handler,
get child loggers, and retrieve the root logger. It also provides constants for different logging levels.
"""

from logging import CRITICAL, DEBUG, ERROR, INFO, NOTSET, WARNING

from ._logging import (
    catch_default_handler,
    disable_default_handler,
    enable_default_handler,
    get_child_logger,
    get_handler,
    get_root_logger,
)

__all__ = (
    "catch_default_handler",
    "disable_default_handler",
    "enable_default_handler",
    "get_child_logger",
    "get_handler",
    "get_root_logger",
    "CRITICAL",
    "DEBUG",
    "ERROR",
    "INFO",
    "NOTSET",
    "WARNING",
)



import os
import subprocess
from pathlib import Path
from typing import Optional, Union

from packing_packages.constants import (
    EXTENSIONS_CONDA,
    EXTENSIONS_PYPI,
)
from packing_packages.utils import check_encoding, is_installed

if is_installed("tqdm"):
    from tqdm import tqdm
else:
    from packing_packages.utils import (  # type: ignore[assignment]
        dummy_tqdm as tqdm,
    )

from packing_packages.logging import get_child_logger

_logger = get_child_logger(__name__)


def install_packages(
    env_name: Optional[str] = None,
    dirpath_packages: Union[str, os.PathLike] = ".",
    encoding: Optional[str] = None,
) -> None:
    """install conda packages

    Parameters
    ----------
    env_name : str, optional
        conda environment name, by default None
    dirpath_packages : Union[str, os.PathLike], optional
        directory path of packages
    """
    encoding = check_encoding(encoding)

    if env_name is None:
        env_name = os.environ["CONDA_DEFAULT_ENV"]
    else:
        # check environment name
        result_conda_env_list = subprocess.run(
            [
                os.environ["CONDA_EXE"],
                "info",
                "-e",
            ],
            stdout=subprocess.PIPE,
        )
        conda_env_list = result_conda_env_list.stdout.decode(
            encoding
        ).splitlines()
        env_name_list = {
            line.split()[0]
            for line in conda_env_list
            if line and line[0] != "#"
        }
        if env_name not in env_name_list:
            raise ValueError(f"Environment '{env_name}' not found.")

    _logger.info(
        f"Installing packages from '{dirpath_packages}' into '{env_name}'"
    )

    dirpath_packages = Path(dirpath_packages).resolve()

    list_filepaths_conda: list[Path] = sum(
        [
            [
                filepath
                for filepath in dirpath_packages.glob(f"**/*{ext}")
                if filepath.is_file()
            ]
            for ext in EXTENSIONS_CONDA
        ],
        start=[],
    )
    list_filepaths_pypi: list[Path] = sum(
        [
            [
                filepath
                for filepath in dirpath_packages.glob(f"**/*{ext}")
                if filepath.is_file()
            ]
            for ext in EXTENSIONS_PYPI
        ],
        start=[],
    )

    list_filepaths_conda_failed: list[Path] = []
    # install conda packages
    for filepath_conda in tqdm(
        list_filepaths_conda,
        desc="Installing conda packages",
        unit="package",
    ):
        result_conda_install = subprocess.run(
            [
                os.environ["CONDA_EXE"],
                "install",
                "-y",
                "-n",
                env_name,
                "--offline",
                "--use-local",
                str(filepath_conda),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _logger.info(result_conda_install.stdout.decode(encoding))
        if result_conda_install.returncode != 0:
            _logger.warning(
                f"Failed to install conda package: {filepath_conda}"
            )
            _logger.error(result_conda_install.stderr.decode(encoding))
            list_filepaths_conda_failed.append(filepath_conda)

    # install pypi packages
    list_filepaths_pypi_failed: list[Path] = []
    for filepath_pypi in tqdm(
        list_filepaths_pypi,
        desc="Installing PyPI packages",
        unit="package",
    ):
        result_pip_install = subprocess.run(
            [
                os.environ["CONDA_EXE"],
                "run",
                "-n",
                env_name,
                "pip",
                "install",
                "--no-deps",
                "--no-build-isolation",
                str(filepath_pypi),
            ],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        _logger.info(result_pip_install.stdout.decode(encoding))
        if result_pip_install.returncode != 0:
            _logger.warning(f"Failed to install PyPI package: {filepath_pypi}")
            _logger.error(result_pip_install.stderr.decode(encoding))
            list_filepaths_pypi_failed.append(filepath_pypi)

    if list_filepaths_conda_failed:
        _logger.warning(
            f"Failed to install {len(list_filepaths_conda_failed)} conda packages."
        )
        for filepath in list_filepaths_conda_failed:
            _logger.warning(f"  - {filepath}")
    else:
        _logger.info("All conda packages installed successfully.")
    if list_filepaths_pypi_failed:
        _logger.warning(
            f"Failed to install {len(list_filepaths_pypi_failed)} PyPI packages."
        )
        for filepath in list_filepaths_pypi_failed:
            _logger.warning(f"  - {filepath}")
    else:
        _logger.info("All PyPI packages installed successfully.")

    _logger.info(
        f"Installed {len(list_filepaths_conda) + len(list_filepaths_pypi)} packages."
    )
    _logger.info(
        f"Failed to install {len(list_filepaths_conda_failed) + len(list_filepaths_pypi_failed)} packages."
    )
    if not list_filepaths_conda_failed and not list_filepaths_pypi_failed:
        _logger.info("All packages installed successfully.")
    else:
        _logger.warning(
            "Some packages failed to install. Please check the logs."
        )



"""Packing packages for distribution and installation.

This module provides functionality to pack conda environments and their
dependencies into a specified directory.

.. code-block:: bash

    $ python -m packing_packages pack --help

    or

    $ packing-packages pack --help

"""

from ._core import packing_packages

__all__ = ("packing_packages",)



"""utils module for packing_packages.

This module provides utility functions and classes for the packing_packages package.
It includes functions for checking if a package is installed, creating dummy progress bars,
and checking if an argument is passed to a function.
"""

from ._utils import check_encoding, dummy_tqdm, is_argument, is_installed

# _utils.pyだと、_が入っているのでドキュメント化されない。
# ドキュメント化したい場合は、モジュールメソッドとして登録するため、__all__に入れる。
__all__ = ("check_encoding", "dummy_tqdm", "is_argument", "is_installed")



import os
import subprocess
import sys
from pathlib import Path
from shutil import copyfile
from typing import NamedTuple, Optional, Union

from packing_packages.constants import EXTENSIONS_CONDA
from packing_packages.logging import get_child_logger
from packing_packages.utils import check_encoding, is_installed

if is_installed("tqdm"):
    from tqdm.auto import tqdm
else:
    from packing_packages.utils import (  # type: ignore[assignment]
        dummy_tqdm as tqdm,
    )


class Package(NamedTuple):
    name: str
    version: str
    build: str
    channel: str


_logger = get_child_logger(__name__)


def packing_packages(
    env_name: Optional[str] = None,
    dirpath_target: Union[os.PathLike, str] = ".",
    encoding: Optional[str] = None,
    dry_run: bool = False,
) -> None:
    """packaging conda environment packages

    Parameters
    ----------
    env_name : str
        conda environment name
    dirpath_target : str
        target directory path
    dry_run : bool
        if True, do not download files
    """

    encoding = check_encoding(encoding)

    if dry_run:
        _logger.warning("This is a dry run. No files will be downloaded.")

    # check install conda
    if "CONDA_EXE" not in os.environ:
        raise ValueError(
            "Please install conda and set the CONDA_EXE environment variable."
        )

    DIRPATH_CONDA_ROOT = Path(os.environ["CONDA_EXE"]).parent.parent.resolve()
    dirpath_pkgs = DIRPATH_CONDA_ROOT / "pkgs"

    if env_name is None:
        env_name = os.environ["CONDA_DEFAULT_ENV"]
    else:
        # check environment name
        result_conda_env_list = subprocess.run(
            [
                os.environ["CONDA_EXE"],
                "info",
                "-e",
            ],
            stdout=subprocess.PIPE,
        )
        conda_env_list = result_conda_env_list.stdout.decode(
            encoding
        ).splitlines()
        env_name_list = {
            line.split()[0]
            for line in conda_env_list
            if line and line[0] != "#"
        }
        if env_name not in env_name_list:
            raise ValueError(
                f"'{env_name}' is not found. Please check the environment name."
            )
    _logger.info(f"Packing conda environment '{env_name}'...")

    dirpath_target = Path(dirpath_target).resolve()
    if not dirpath_target.is_dir():
        raise FileNotFoundError(dirpath_target)

    dirpath_output = dirpath_target / env_name
    if not dry_run:
        os.makedirs(dirpath_output, exist_ok=True)

    _logger.info(f"Packing to '{dirpath_output}'...")

    dirpath_output_pypi = dirpath_output / "pypi"
    dirpath_output_conda = dirpath_output / "conda"
    for _dirpath in (dirpath_output_pypi, dirpath_output_conda):
        if not dry_run:
            os.makedirs(_dirpath, exist_ok=True)

    result_conda_list = subprocess.run(
        [
            os.environ["CONDA_EXE"],
            "list",
            "-n",
            env_name,
        ],
        stdout=subprocess.PIPE,
    )
    conda_list = result_conda_list.stdout.decode(encoding).splitlines()

    # conda listの出力をパースする
    list_packages: list[Package] = []
    for line in conda_list:
        if line[0] == "#":
            continue
        line_split = tuple(line.strip().split())
        package = Package(
            name=line_split[0],
            version=line_split[1],
            build=line_split[2],
            channel=line_split[3] if len(line_split) > 3 else "",
        )
        list_packages.append(package)
        if package.name == "python":
            env_python_version = package.version

    # # conda listで取得したpythonのバージョンと、現在のpythonのバージョンが異なる場合は警告を出す
    if (
        tuple(map(int, env_python_version.split(".")[:2]))
        != sys.version_info[:2]
    ):
        _logger.warning(
            "The Python version of the conda environment and the current Python version do not match. "
            "It is recommended to use the same Python version."
        )

    list_failed_packages: list[Package] = []
    for package in tqdm(list_packages):
        # pypiの場合はダウンロードする
        if package.channel == "pypi":
            if dry_run:
                result_pip_download = subprocess.run(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        f"{package.name}=={package.version}",
                        "--no-deps",
                        "--no-build-isolation",
                        "--python-version",
                        f"{env_python_version}",
                        "--dry-run",
                        "--target",
                        dirpath_output_pypi.as_posix(),
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            else:
                result_pip_download = subprocess.run(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "download",
                        f"{package.name}=={package.version}",
                        "--no-deps",
                        "--python-version",
                        f"{env_python_version}",
                        "-d",
                        dirpath_output_pypi.as_posix(),
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

            _logger.info("\n" + result_pip_download.stdout.decode(encoding))
            if result_pip_download.returncode != 0:
                _logger.warning(f"'{package}' is not found.")
                _logger.error(result_pip_download.stderr.decode(encoding))
                list_failed_packages.append(package)

            continue

        # conda
        tup_filepaths_package = tuple(
            dirpath_pkgs
            / f"{package.name}-{package.version}-{package.build}.{ext}"
            for ext in EXTENSIONS_CONDA
        )
        # ないなら、conda installでダウンロードする
        if not any(
            filepath_package.is_file()
            for filepath_package in tup_filepaths_package
        ):
            result_conda_download = subprocess.run(
                [
                    os.environ["CONDA_EXE"],
                    "install",
                    "-y",
                    "-n",
                    env_name,
                    f"{package.name}={package.version}={package.build}",
                    "--no-deps",
                    "--download-only",
                ]
                + (["-c", package.channel] if package.channel else [])
                + (["--dry-run"] if dry_run else []),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            _logger.info(result_conda_download.stdout.decode(encoding))
            if result_conda_download.returncode != 0:
                _logger.warning(f"'{package}' is not found.")
                _logger.error(result_conda_download.stderr.decode(encoding))
                list_failed_packages.append(package)
                # 失敗したら次のパッケージへ (コピーできない)
                continue

        # ある or donwloadに成功したときはコピーする
        for filepath_package in tup_filepaths_package:
            if filepath_package.is_file():
                _logger.info(f"copying {filepath_package.name}")
                if not dry_run:
                    copyfile(
                        filepath_package,
                        dirpath_output_conda / filepath_package.name,
                    )
                break

    n_success = len(list_packages) - len(list_failed_packages)
    n_failed = len(list_failed_packages)
    _logger.info(
        f"{n_success} / {(n_success + n_failed)} packages are success!"
    )

    if n_failed > 0:
        for package in list_failed_packages:
            _logger.warning(f"  {package.name}=={package.version}")
        _logger.info(
            "Please check the above packages. "
            "If you want to install them, please run the following command:"
        )


if __name__ == "__main__":
    packing_packages("open-webui", "./open-webui")
